
from twisted.trial import unittest
import config_manager
import os
import logging
import shutil
from StringIO import StringIO

import pqm

sample_message = ("From: John.Citizen@example.com\n"
                  "Subject: A new action\n"
                  "star-merge archive@example.com/foo--bar--0 "
                  "archive2@example.com/bar--foo--1")
sample_message_2 = ("From: John.Citizen@example.com\n"
                    "Subject: A new action\n"
                    "star-merge archive@example.com/foo--bar--0 "
                    "archive2@example.com/bar--foo--2")
sample_message_bad = ("From: John.Citizen@example.com\n"
                      "Subject: A new action\n"
                      "star-merge archive@example.com/foo--bar--0 "
                      "unregistered@example.com/bar--foo--1")
sample_signed_message = ("From: whee@bar.com (Matthew Thomas)\n"
    "Subject: [trivial] fix various actions portlet icons\n"
    "-----BEGIN PGP SIGNED MESSAGE-----\n"
    "Hash: SHA1\n"
    "\n"
    "star-merge mpt@canonical.com/launchpad--devel--0 rocketfuel@canonical.com/launchpad--devel--0\n"
    "-----BEGIN PGP SIGNATURE-----\n"
    "Version: GnuPG v1.2.5 (GNU/Linux)\n"
    "\n"
    "iD8DBQFC8guw6PUxNfU6ecoRAt5OAKDPyd1Ac0H7QO4DIEiZ2wvscqrg0ACeNL0h\n"
    "ksM1dJeD/aX54NnE/uqps/M=\n"
    "=X7Ib\n"
    "-----END PGP SIGNATURE-----\n"
    "\n")

sample_config_with_overrides = ("[DEFAULT]\n"
                                "queuedir=/tmp\n"
                                "[location overrides]\n"
                                "/local/path=sftp://host/otherpath\n"
                                "scheme://something=righthandside\n")

sample_config_default_project = ("[DEFAULT]\n"
                                "queuedir=/tmp\n"
                                "project=default-project\n"
                                "[/dummy/]\n"
                                )

sample_config_two_project_specs = ("[DEFAULT]\n"
                                   "queuedir=/tmp\n"
                                   "[/one/]\n"
                                   "project=project-one\n"
                                   "[/two/]\n"
                                   "project=project-two\n"
                                   "[/three/]\n"
                                   "project=project-two\n")

sample_config_default_and_spec_projects = (
                                   "[DEFAULT]\n"
                                   "queuedir=/tmp\n"
                                   "project=default-project\n"
                                   "[/one/]\n"
                                   "project=project-one\n"
                                   "[/two/]\n"
                                   "project=project-two\n"
                                   "[/three/]\n"
                                   "project=project-two\n")

sample_config_unpublished = ("[DEFAULT]\n"
                             "queuedir=/tmp\n"
                             "[foo@bar/gam--foo--0]\n")

sample_config_published = ("[DEFAULT]\n"
                           "queuedir=/tmp\n"
                           "[foo@bar/gam--foo--0]\n"
                           "published_at=file:///a\n"
                           "[file:///tmp/foo]\n"
                           "published_at=file:///b\n"
                           "publish_to=sftp://host/b\n"
                           "[file:///tmp/bar/]\n"
                           "published_at=file:///c/\n"
                           )

sample_config_with_groups = ("[DEFAULT]\n"
                             "queuedir=/tmp\n"
                             "groups=agroup\n"
                             "[agroup]\n"
                             "members=foo@bar.com\n"
                             )

sample_config_with_workdir = ("[DEFAULT]\n"
                              "queuedir=queue\n"
                              "workdir=test-workdir\n"
                              )


class QueueSetup(object):
    """Setup a queue with mock messages in it."""

    def __init__(self):
        self.configFileName = "Foo"
        self.cwd = os.path.abspath(os.path.curdir)
        self.message = sample_message
        self.message_bad = sample_message_bad
        self.message3 = sample_message_2

    def setUp(self):
        myFile=open(self.configFileName, "w")
        myFile.write("[DEFAULT]\n"
                     "queuedir=%s/queue\n"
                     "[archive2@example.com/bar--foo--1]\n"
                     "project=project\n"
                     "[archive2@example.com/bar--foo--2]\n"
                     % self.cwd)
        myFile.close()
        self.queuedir = os.path.join(self.cwd, "queue")
        try:
            os.mkdir(self.queuedir)
            os.mkdir(self.queuedir + '/pqm')
        except OSError:
            pass
        self.messageFileName = os.path.join(self.queuedir, "patch.00000001")
        messageFile = open(self.messageFileName, "w")
        messageFile.write(self.message)
        messageFile.close()
        self.messageFileNameBad = os.path.join(self.queuedir, "patch.00000002")
        messageFile = open(self.messageFileNameBad, "w")
        messageFile.write(self.message_bad)
        messageFile.close()
        self.messageFileName3 = os.path.join(self.queuedir, "patch.00000003")
        messageFile = open(self.messageFileName3, "w")
        messageFile.write(self.message3)
        messageFile.close()

    def tearDown(self):
        os.unlink(self.configFileName)
        shutil.rmtree(self.queuedir, ignore_errors=True)


class TestWithQueueDirectory(unittest.TestCase):
    
    def setUp(self):
        unittest.TestCase.setUp(self)
        self.queue = QueueSetup()
        self.queue.setUp()
    
    def testName(self):
        patch = pqm.Script('foo.script', logging, False, 0, None, None)
        self.assertEqual(patch.filename, 'foo.script')
        self.scriptname = 'fpp'

    def tearDown(self):
        try:
            os.unlink(self.scriptname)
        except OSError:
            pass
        self.queue.tearDown()

    def getScript(self, content):
        """Get a script for testing with."""
        self.scriptname = 'foo.script'
        scriptFile = open(self.scriptname, "w")
        scriptFile.write(content)
        scriptFile.close()
        pqm.pqm_subdir = os.path.abspath('queue/pqm')
        configp = pqm.ConfigParser()
        configp.read([self.queue.configFileName])
        handler = pqm.BranchSpecOptionHandler(configp)
        return pqm.Script(self.scriptname, logging, False, 54, handler, configp)

    def testFields(self):
        script = self.getScript(sample_message)
        self.assertEqual(script.getSender(), "John.Citizen@example.com")
        self.assertEqual(script.getSubject(), "A new action")
        self.assertEqual(script.getContent(), 
            ("star-merge archive@example.com/foo--bar--0 "
             "archive2@example.com/bar--foo--1"))
        self.assertEqual(script.getLines(), 
            [("star-merge archive@example.com/foo--bar--0 "
             "archive2@example.com/bar--foo--1")])
        self.assertEqual([pqm.MergeCommand(None,
                                           None,
                                           None,
                                           'archive@example.com/foo--bar--0',
                                           'archive2@example.com/bar--foo--1')],
                         script.getCommands())

    def testGPGFields(self):
        script = self.getScript(sample_signed_message)
        self.assertEqual(script.getSender(), "whee@bar.com (Matthew Thomas)")
        self.assertEqual(script.getSubject(), 
            "[trivial] fix various actions portlet icons")
        self.assertEqual(script.getContent(), 
            ('-----BEGIN PGP SIGNED MESSAGE-----\n' 
             'Hash: SHA1\n' 
             '\n' 
             'star-merge mpt@canonical.com/launchpad--devel--0 '
             'rocketfuel@canonical.com/launchpad--devel--0\n'
             '-----BEGIN PGP SIGNATURE-----\n' 
             'Version: GnuPG v1.2.5 (GNU/Linux)\n'
             '\n'
             'iD8DBQFC8guw6PUxNfU6ecoRAt5OAKDPyd1Ac0H7QO4DIEiZ2wvscqrg0ACeNL0h\n'
             'ksM1dJeD/aX54NnE/uqps/M=\n'
             '=X7Ib\n'
             '-----END PGP SIGNATURE-----\n'
             '\n'))
        self.assertEqual(script.getLines(), 
            [("star-merge mpt@canonical.com/launchpad--devel--0 "
             "rocketfuel@canonical.com/launchpad--devel--0")])
        self.assertEqual(
            [pqm.MergeCommand(None,
                              None,
                              None,
                              'mpt@canonical.com/launchpad--devel--0',
                              'rocketfuel@canonical.com/launchpad--devel--0')],
            script.getCommands())

    def testDate(self):
        """Can we access a submission time for scripts."""
        script = self.getScript(sample_message)
        self.assertEqual(script.getSubmissionTime(), 54)

    def testProjects(self):
        # A script may affect multiple branches, so we can ask the script
        # for its projects, and then for the commands for a project.
        script = self.getScript(sample_message)
        self.assertEqual(set(['project']), script.getProjects())

    def test_log_status(self):
        # make an initial status file
        status = file(self.queue.queuedir + '/pqm/status', 'wt')
        status.write('bar\n')
        status.close()
        script = self.getScript(sample_message)
        command = pqm.Command(script,
                              script._branch_spec_handler,
                              script._configp)
        logger = MockLogger()
        command.log_with_status(logger, 'foo')
        content = file(self.queue.queuedir + '/pqm/status').read()
        self.assertEqual('bar\nfoo\n', content)



# NOTDONEYET: move command recognition from CommandRunner to Script
# and commands should follow the command pattern rather than being strings.
#
#    def testIsCommand(self):
#        """Is a line a command line?"""
#        import pqm
#        script = pqm.Script('foo.script', logging, False)
#        self.failIf(script.is_command(""))


class MockLogger(object):

    def info(self, message, *args, **kwargs):
        """Pretend to work."""
        pass


class MockScript(object):

    def __init__(self):
        """Create a Script which is only useful for testing."""
        self.debug = False
        self.signing_email = 'gpg_sender'

    def getSubject(self):
        return "a subject."

    def getSender(self):
        return "a_sender"


class TestCommandRunner(unittest.TestCase):

    def test_star_merge_urls(self):
        from pqm import Script
        star_match = Script.star_re.match("star-merge file:///url1 file:///url2")
        self.assertEqual(star_match.group(1), 'file:///url1')
        self.assertEqual(star_match.group(2), 'file:///url2')
        
    def test_star_merge_names(self):
        from pqm import Script
        star_match = Script.star_re.match("star-merge foo@bar/foo--bar--0 foo@bar/gam--foo--0")
        self.assertEqual(star_match.group(1), 'foo@bar/foo--bar--0')
        self.assertEqual(star_match.group(2), 'foo@bar/gam--foo--0')

    def test_get_arch_impl(self):
        from pqm import CommandRunner
        configp = pqm.ConfigParser()
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = CommandRunner(None, handler, configp, [])
        # TODO should be able to pass on systems without baz installed.
        # TODO set a fake logger
        self.failUnless(isinstance(runner.get_arch_impl(),
                                   pqm.Baz1_1Handler))

    def test_check_revision(self):
        configp = pqm.ConfigParser()
        handler = pqm.BranchSpecOptionHandler(configp)
        handler._specs = {'foo@bar/gam--foo--0':{},
                          'file:///tmp/foo':{},
                          'file:///tmp/bar/':{}}
        runner = pqm.CommandRunner(None, handler, configp, [])
        runner.script = MockScript()
        runner.check_target('foo@bar/gam--foo--0', 'blah')
        runner.check_target('file:///tmp/foo', 'blah')
        self.assertRaises(pqm.PQMCmdFailure,
                          runner.check_target,
                          'file:///tmp/foo/branch',
                          'blah')
        self.assertRaises(pqm.PQMCmdFailure,
                          runner.check_target,
                          'file:///tmp/bar',
                          'blah')
        runner.check_target('file:///tmp/bar/branch', 'blah')
 
    def test__make_wd_path(self):
        configp = pqm.ConfigParser()
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = pqm.CommandRunner(None, handler, configp, [])
        self.assertEqual(runner._make_wd_path('/foo', 'foo@bar/gam--foo--0'),
                         '/foo/foo@bar')
        self.assertEqual(runner._make_wd_path('/foo', '/foo/bar'), '/foo/foo')
        self.assertEqual(runner._make_wd_path('/foo', 'file:///foo/bar'), 
                         '/foo/file')
        self.assertEqual(runner._make_wd_path('/foo', 'branch'), '/foo/branch')

    def test__branch_name(self):
        configp = pqm.ConfigParser()
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = pqm.CommandRunner(None, handler, configp, [])
        self.assertEqual(runner._branch_name('foo@bar/gam--foo--0'),
                         'gam--foo--0')
        self.assertEqual(runner._branch_name('foo'), 'foo')
        self.assertEqual(runner._branch_name('file:///home/bar/foo'), 'foo')
        self.assertEqual(runner._branch_name('file:///home/bar/foo'), 'foo')

    def test_cleanup_wd_missing_dir(self):
        configp = pqm.ConfigParser()
        configp.readfp(StringIO(sample_config_with_workdir))
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = pqm.CommandRunner(None, handler, configp, [])
        pqm.workdir = "test-workdir"
        pqm.branch_specs={'path':{'build_dir':'nonexistant-dir'}}
        os.mkdir("test-workdir")
        try:
            runner.cleanup_wd()
        finally:
            shutil.rmtree("test-workdir")

    def test__get_url_override_mapper_no_overrides(self):
        configp = pqm.ConfigParser()
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = pqm.CommandRunner(None, handler, configp, [])
        runner.configp = pqm.ConfigParser()
        expected_mapper = config_manager.URLMapper()
        self.assertEqual(expected_mapper, runner._get_url_override_mapper())
        
    def test__get_url_override_mapper(self):
        configp = pqm.ConfigParser()
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = pqm.CommandRunner(None, handler, configp, [])
        runner.configp = pqm.ConfigParser()
        runner.configp.readfp(StringIO(sample_config_with_overrides))
        expected_mapper = config_manager.URLMapper()
        expected_mapper.add_map("sftp://host/otherpath", "/local/path")
        expected_mapper.add_map("righthandside", "scheme://something")
        self.assertEqual(expected_mapper, runner._get_url_override_mapper())
        

class TestCommand(unittest.TestCase):

    def test_debug_command_sets_script_debug(self):
        configp = pqm.ConfigParser()
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)
        script = MockScript()
        self.assertFalse(script.debug)
        command = pqm.DebugCommand(script, handler, configp)
        successes, unrecognized, output = command.run()
        self.assertEqual(['debug'], successes)
        self.assertEqual([], unrecognized)
        self.assertEqual(['Debug mode enabled.'], output)
        self.assertTrue(script.debug)


class FunctionalTestCommandRunner(unittest.TestCase):

    def setUp(self):
        from bzrlib.add import smart_add_tree
        from bzrlib.bzrdir import BzrDir
        from bzrlib.plugin import load_plugins
        load_plugins()
        tree = BzrDir.create_standalone_workingtree("bzrbranch")
        smart_add_tree(tree, ["bzrbranch"])
        tree.commit("start branch.", verbose=False)

    def tearDown(self):
        shutil.rmtree("bzrbranch")

    def test_get_branch_handler_arch(self):
        # I would test this, but I don't want to add a full dependency on 
        # pybaz at this point
        #from pqm import CommandRunner
        #runner = CommandRunner()
        ## TODO set a fake logger
        pass
        
    def branch_url(self):
        branchurl = "file://%s" % os.path.abspath("bzrbranch")
        return branchurl.replace('\\', '/')
        
    def test_get_branch_handler_bzr_file(self):
        configp = pqm.ConfigParser()
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = pqm.CommandRunner(None, handler, configp, [])
        runner.script = MockScript()
        self.failUnless(isinstance(runner.get_branch_handler(self.branch_url()),
                                   pqm.Bazaar2Handler))
        
    def test_get_branch_handler_bzr_relative(self):
        configp = pqm.ConfigParser()
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = pqm.CommandRunner(None, handler, configp, [])
        runner.script = MockScript()
        self.failUnless(isinstance(runner.get_branch_handler("bzrbranch"),
                                   pqm.Bazaar2Handler))
 
    def test_set_current_vcs(self):
        configp = pqm.ConfigParser()
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)
        runner = pqm.CommandRunner(None, handler, configp, [])
        runner.script = MockScript()
        self.assertEqual(runner.get_vcs(), None)
        runner.set_current_vcs(self.branch_url(), "bzrbranch")
        self.assertNotEqual(runner.get_vcs(), None)
        self.failUnless(isinstance(runner.get_vcs(), pqm.Bazaar2Handler))


class BzrHandlerTestCase(unittest.TestCase):

    def setUp(self):
        from bzrlib.add import smart_add_tree
        from bzrlib.bzrdir import BzrDir
        from bzrlib.plugin import load_plugins
        load_plugins()
        tree = BzrDir.create_standalone_workingtree("bzrbranch")
        smart_add_tree(tree, ["bzrbranch"])
        tree.commit("start branch.", verbose=False)
        tree.branch.bzrdir.sprout("bzrbranch-parent")
        tree.branch.bzrdir.sprout("bzrbranch-public")
        tree.branch.bzrdir.sprout("branch-contributor")
        contrib_tree = BzrDir.open("branch-contributor").open_workingtree()
        file = open("branch-contributor/README", "wt")
        file.write("Boo!\n")
        file.close()
        contrib_tree.add(["README"])
        contrib_tree.commit("add README")

    def tearDown(self):
        shutil.rmtree("bzrbranch")
        shutil.rmtree("branch-contributor")
        shutil.rmtree("bzrbranch-parent")
        shutil.rmtree("bzrbranch-public")

    def test_commit(self):
        from bzrlib.branch import Branch
        branch = Branch.open("bzrbranch")
        self.assertEqual(branch.repository.get_revision(branch.last_revision()).message,
                         'start branch.')
        handler = pqm.Bazaar2Handler()
        config = {'publish_to':'bzrbranch-public'}
        handler.commit("me", "bzrbranch", "wheee", "bzrbranch-parent", config)
        for path in ['bzrbranch', 'bzrbranch-parent', 'bzrbranch-public']:
            branch = Branch.open(path)
            rev = branch.last_revision()
            self.assertEqual(branch.repository.get_revision(rev).message, 'wheee')

    def test_commit_uncommits_when_push_fails(self):
        from bzrlib.branch import Branch
        branch = Branch.open("bzrbranch")
        self.assertEqual(branch.repository.get_revision(branch.last_revision()).message,
                         'start branch.')
        handler = pqm.Bazaar2Handler()
        config = {'publish_to':'bzrbranch-public-missing'}
        self.assertRaises(pqm.PQMTlaFailure, handler.commit, "me",
                          "bzrbranch", "wheee", "bzrbranch-parent", config)
        # bzr branch's commit is allowed as its the local tree - we ont
        # care what happens there, but the
        # parent and published copies must be reverted.
        for path in ['bzrbranch-parent', 'bzrbranch-public']:
            branch = Branch.open(path)
            rev = branch.last_revision()
            self.assertEqual(branch.repository.get_revision(rev).message, 'start branch.')

    def test_make_local_dir(self):
        try:
            handler = pqm.Bazaar2Handler()
            handler.make_local_dir("me", "bzrbranch", "proof")
            from bzrlib.branch import Branch
            branch = Branch.open("proof")
            self.assertEqual(branch.repository.get_revision(branch.last_revision()).message,
                             'start branch.')
        finally:
            shutil.rmtree("proof")
            
    def test_merge(self):
        from bzrlib.branch import Branch
        branch = Branch.open("bzrbranch")
        self.assertEqual(branch.repository.get_revision(branch.last_revision()).message,
                         'start branch.')
        contrib_branch = Branch.open("branch-contributor")
        self.assertEqual(contrib_branch.repository.get_revision(
                            contrib_branch.last_revision()).message,
                         'add README')
        self.failIf(os.path.exists("bzrbranch/README"))
        handler = pqm.Bazaar2Handler()
        result = handler.do_star_merge("me", "branch-contributor", "bzrbranch")
        self.assertEqual(result, ["merge successful"])
        branch = Branch.open("bzrbranch")
        self.assertEqual(branch.repository.get_revision(branch.last_revision()).message,
                         'start branch.')
        contrib_branch = Branch.open("branch-contributor")
        self.assertEqual(contrib_branch.repository.get_revision(
                            contrib_branch.last_revision()).message,
                         'add README')
        self.failUnless(os.path.exists("bzrbranch/README"))
        
    def test_merge_conflicts(self):
        from bzrlib.workingtree import WorkingTree
        from bzrlib.branch import Branch
        tree = WorkingTree.open('bzrbranch')
        self.assertEqual(tree.branch.repository.get_revision(tree.last_revision()).message,
                         'start branch.')
        contrib_branch = Branch.open("branch-contributor")
        self.assertEqual(contrib_branch.repository.get_revision(
                            contrib_branch.last_revision()).message,
                         'add README')
        contrib_tree = WorkingTree.open('branch-contributor')
        # give tree a README with common ancestry
        tree.pull(contrib_branch)
        # commit new READMEs to both trees.
        file = open("bzrbranch/README", "wt")
        file.write("Boo conflict!\n")
        file.close()
        tree.commit('add conflicting README')
        
        file = open("branch-contributor/README", "wt")
        file.write("Boo conflict YEAH!\n")
        file.close()
        contrib_tree.commit('add conflicting README')
        handler = pqm.Bazaar2Handler()
        e = None
        try:
            handler.do_star_merge("me", "branch-contributor", "bzrbranch")
        except pqm.PQMTlaFailure, e:
            self.assertEqual(["Conflicts during merge: ",
                              "Text conflict in README",
                              ],
                             e.output)
            return
        self.fail("Merge conflict not raised.")


class TestConfig(unittest.TestCase):

    def test_location_overrides(self):
        config = pqm.ConfigParser()
        config.readfp(StringIO(sample_config_with_overrides))
        location_rules = config.section_items('location overrides')
        expected_rules = [("/local/path", "sftp://host/otherpath"),
                          ("scheme", "//something=righthandside")]
        self.assertEqual(set(expected_rules), set(location_rules))

    def test_projects_empty(self):
        # empty config -> no projects
        configp = pqm.ConfigParser()
        pqm.pqm_subdir = '/tmp' # ewww
        branch_spec_handler = pqm.BranchSpecOptionHandler(configp)
        self.assertEqual(set(), branch_spec_handler.project_names())

    def test_projects_default(self):
        # default project -> its listed and present as a branch option.
        configp = pqm.ConfigParser()
        configp.readfp(StringIO(sample_config_default_project))
        pqm.pqm_subdir = '/tmp' # ewww
        branch_spec_handler = pqm.BranchSpecOptionHandler(configp)
        self.assertEqual(set(['default-project']),
                         branch_spec_handler.project_names())
        self.assertEqual('default-project',
            branch_spec_handler.get_target_config('/dummy/foo')[1]['project'])
        
    def test_projects_branch_specs(self):
        # branch specs have their projects aggregated and dont error with
        # one or two specs mentionining a project.
        configp = pqm.ConfigParser()
        configp.readfp(StringIO(sample_config_two_project_specs))
        pqm.pqm_subdir = '/tmp' # ewww
        branch_spec_handler = pqm.BranchSpecOptionHandler(configp)
        self.assertEqual(set(['project-one', 'project-two']),
                         branch_spec_handler.project_names())
        self.assertEqual('project-one',
            branch_spec_handler.get_target_config('/one/foo')[1]['project'])
        self.assertEqual('project-two',
            branch_spec_handler.get_target_config('/two/foo')[1]['project'])
        self.assertEqual('project-two',
            branch_spec_handler.get_target_config('/two/foo')[1]['project'])
        
    def test_projects_branch_specs_and_default(self):
        # branch specs have their projects aggregated with the default spec.
        configp = pqm.ConfigParser()
        configp.readfp(StringIO(sample_config_default_and_spec_projects))
        pqm.pqm_subdir = '/tmp' # ewww
        branch_spec_handler = pqm.BranchSpecOptionHandler(configp)
        self.assertEqual(set(['default-project', 'project-one', 'project-two']),
                         branch_spec_handler.project_names())
        self.assertEqual('project-one',
            branch_spec_handler.get_target_config('/one/foo')[1]['project'])
        self.assertEqual('project-two',
            branch_spec_handler.get_target_config('/two/foo')[1]['project'])
        self.assertEqual('project-two',
            branch_spec_handler.get_target_config('/two/foo')[1]['project'])
        
    def test_get_target_unpublished_config(self):
        configp = pqm.ConfigParser()
        configp.readfp(StringIO(sample_config_unpublished))
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)
        branch, config = handler.get_target_config('foo@bar/gam--foo--0')
        self.assertEqual('foo@bar/gam--foo--0', branch)
        self.assertEqual(None, config['published_at'])
        self.assertEqual(None, config['publish_to'])
        self.assertRaises(KeyError, handler.get_target_config, 'foo@bar/gam--0')
        
    def test_get_target_config(self):
        configp = pqm.ConfigParser()
        configp.readfp(StringIO(sample_config_published))
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)
        branch, config = handler.get_target_config('foo@bar/gam--foo--0')
        self.assertEqual('foo@bar/gam--foo--0', branch)
        self.assertEqual('file:///a', config['published_at'])
        self.assertEqual('file:///a', config['publish_to'])
        # close but shouldn't match:
        self.assertRaises(KeyError, handler.get_target_config, 'foo@bar/gam--0')
        self.assertRaises(KeyError, handler.get_target_config, 'file:///a/bar')
        # note that publish_to is not overriden by the default from published_at.
        branch, config = handler.get_target_config('file:///tmp/foo')
        self.assertEqual('file:///tmp/foo', branch)
        self.assertEqual('file:///b', config['published_at'])
        self.assertEqual('sftp://host/b', config['publish_to'])
        # get published_at -> real location, publish_to still sftp://
        branch, config = handler.get_target_config('file:///b')
        self.assertEqual('file:///tmp/foo', branch)
        self.assertEqual('file:///b', config['published_at'])
        self.assertEqual('sftp://host/b', config['publish_to'])
        # close but shouldn't match:
        self.assertRaises(KeyError,
                          handler.get_target_config,
                          'file:///tmp/foo/branch')
        self.assertRaises(KeyError,
                          handler.get_target_config,
                          'file:///b/branch')
        branch, config = handler.get_target_config('file:///tmp/bar/branch')
        self.assertEqual('file:///tmp/bar/branch', branch)
        self.assertEqual('file:///c/branch', config['published_at'])
        self.assertEqual('file:///c/branch', config['publish_to'])
        branch, config = handler.get_target_config('file:///c/branch')
        self.assertEqual('file:///tmp/bar/branch', branch)
        self.assertEqual('file:///c/branch', config['published_at'])
        self.assertEqual('file:///c/branch', config['publish_to'])
        # but the location isn't a branch itself
        self.assertRaises(KeyError,
                          handler.get_target_config,
                          'file:///tmp/bar')
        self.assertRaises(KeyError,
                          handler.get_target_config,
                          'file:///c')
        self.assertRaises(KeyError,
                          handler.get_target_config,
                          'file:///c/')

    def test_smoking_groups(self):
        configp = pqm.ConfigParser()
        configp.readfp(StringIO(sample_config_with_groups))
        pqm.pqm_subdir = '/tmp' # ewww
        handler = pqm.BranchSpecOptionHandler(configp)


class TestStatusFile(unittest.TestCase):

    def test_make_new(self):
        status = pqm.StatusFile('statusfile')
        self.assertTrue(os.path.exists('statusfile'))

    def test_line_has_newline(self):
        #status.line always writes a newline
        status = pqm.StatusFile('statusfile')
        status.line('foo')
        content = file('statusfile').read()
        self.assertEqual('foo\n', content)

    def test_truncates_at_5_lines(self):
        statusfile = file('statusfile', 'wt')
        statusfile.write('long line\n')
        statusfile.write('long line\n')
        statusfile.write('long line\n')
        statusfile.write('long line\n')
        statusfile.write('long line\n')
        statusfile.close()
        # after 5 lines the status shows only the end 5 lines
        # even when they are shorter than the previous content.
        status = pqm.StatusFile('statusfile')
        status.line('foo')
        status.line('bar')
        status.line('gam')
        status.line('quux')
        status.line('mary')
        status.line('joseph')
        content = file('statusfile').read()
        self.assertEqual(
            'bar\n'
            'gam\n'
            'quux\n'
            'mary\n'
            'joseph\n',
            content)

    def test_open_does_not_truncate(self):
        statusfile = file('statusfile', 'wt')
        statusfile.write('foo\n')
        statusfile.close()
        status = pqm.StatusFile('statusfile')
        content = file('statusfile').read()
        self.assertEqual('foo\n', content)
